@@ -323,7 +323,7 @@ class PyPIStage(BaseStage):
 
             # we don't have an old result and got a non-404 code.
             raise self.UpstreamError("%s status on GET %s" %
-                                     (response.status_code, url))
+                                     (response.status, url))
 
         # pypi.python.org provides X-PYPI-LAST-SERIAL header in case of 200 returns.
         # devpi-master may provide a 200 but not supply the header
@@ -395,8 +395,7 @@ class PyPIStage(BaseStage):
             threadlog.debug("get_simplelinks pypi: finished waiting for devpi_serial %r",
                             devpi_serial)
             # XXX raise TransactionRestart to get a consistent clean view
-            self.keyfs.commit_transaction_in_thread()
-            self.keyfs.begin_transaction_in_thread()
+            self.keyfs.restart_read_transaction()
             is_fresh, links, cache_serial = self._load_cache_links(project)
             if links is not None:
                 self.cache_link_updates.refresh(project)
