@@ -7,6 +7,7 @@ import subprocess
 
 import pytest
 import py
+import requests
 import socket
 import sys
 from bs4 import BeautifulSoup
@@ -47,12 +48,6 @@ class TimeoutQueue(BaseQueue):
 
 log = threadlog
 
-def pytest_addoption(parser):
-    parser.addoption("--slow", action="store_true", default=False,
-        help="run slow tests involving remote services (pypi.python.org)")
-    parser.addoption("--backend", action="store",
-        help="run tests with specified dotted name backend")
-
 @pytest.fixture(autouse=True)
 def _clear():
     thread_clear_log()
@@ -117,7 +112,7 @@ def auto_transact(request):
         request.node.get_marker("notransaction")):
         yield
         return
-    keyfs = request.getfuncargvalue("keyfs")
+    keyfs = request.getfixturevalue("keyfs")
 
     write = True if request.node.get_marker("writetransaction") else False
     keyfs.begin_transaction_in_thread(write=write)
@@ -258,6 +253,15 @@ def replica_xom(request, makexom):
     return xom
 
 
+@pytest.fixture
+def makefunctionaltestapp(request):
+    def makefunctionaltestapp(host_port):
+        mt = MyFunctionalTestApp(host_port)
+        mt.xom = None
+        return mt
+    return makefunctionaltestapp
+
+
 @pytest.fixture
 def maketestapp(request):
     def maketestapp(xom):
@@ -289,7 +293,7 @@ def make_simple_pkg_info(name, text="", pkgver=None, hash_type=None,
             hash_value = getattr(hashlib, hash_type)(hv).hexdigest()
             ret.hash_spec = "%s=%s" %(hash_type, hash_value)
             pkgver += "#" + ret.hash_spec
-        text = '<a href="../../{name}/{pkgver}" />'.format(name=name, pkgver=pkgver)
+        text = '<a href="../../{name}/{pkgver}">{pkgver}</a>'.format(name=name, pkgver=pkgver)
     elif text and "{md5}" in text:
         text = text.format(md5=getmd5(text))
     elif text and "{sha256}" in text:
@@ -316,6 +320,11 @@ def httpget(pypiurls):
                     xself.allow_redirects = allow_redirects
                     if "content" in fakeresponse:
                         xself.raw = py.io.BytesIO(fakeresponse["content"])
+
+                @property
+                def status(xself):
+                    return "%s" % xself.status_code
+
                 def __repr__(xself):
                     return "<mockresponse %s url=%s>" % (xself.status_code,
                                                          xself.url)
@@ -490,8 +499,12 @@ class Mapp(MappMixin):
         return sorted(links)
 
     def downloadrelease(self, code, url):
-        r = self.testapp.xget(code, url)
-        if code < 300:
+        r = self.testapp.get(url, expect_errors=True)
+        if isinstance(code, tuple):
+            assert r.status_code in code
+        else:
+            assert r.status_code == code
+        if r.status_code < 300:
             return r.body
         return r.json
 
@@ -628,7 +641,7 @@ class Mapp(MappMixin):
         if not isinstance(users, list):
             users = users.split(",")
         assert isinstance(users, list)
-        result["acl_upload"] = users
+        result["acl_" + acltype] = users
         r = self.testapp.patch_json("/%s" % (indexname,), result)
         assert r.status_code == 200
 
@@ -798,7 +811,7 @@ class MyTestApp(TApp):
 
     def push(self, url, params=None, **kw):
         kw.setdefault("expect_errors", True)
-        return self._gen_request("PUSH", url, params=params, **kw)
+        return self._gen_request("POST", url, params=params, **kw)
 
     def get(self, *args, **kwargs):
         kwargs.setdefault("expect_errors", True)
@@ -834,6 +847,53 @@ class MyTestApp(TApp):
         return self.get(*args, **kwargs)
 
 
+class FunctionalResponseWrapper(object):
+    def __init__(self, response):
+        self.res = response
+
+    @property
+    def status_code(self):
+        return self.res.status_code
+
+    @property
+    def body(self):
+        return self.res.content
+
+    @property
+    def json(self):
+        return self.res.json()
+
+
+class MyFunctionalTestApp(MyTestApp):
+    def __init__(self, host_port):
+        import json
+        self.base_url = "http://%s:%s" % host_port
+        self.headers = {}
+        self.JSONEncoder = json.JSONEncoder
+
+    def _gen_request(self, method, url, params=None,
+                     headers=None, extra_environ=None, status=None,
+                     upload_files=None, expect_errors=False,
+                     content_type=None):
+        headers = {} if headers is None else headers.copy()
+        if self.auth:
+            headers["X-Devpi-Auth"] = b64encode("%s:%s" % self.auth)
+
+        # fill headers with defaults
+        for name, val in self.headers.items():
+            headers.setdefault(name, val)
+
+        kw = dict(headers=headers)
+        if params and params is not webtest.utils.NoDefault:
+            if method.lower() in ('post', 'put', 'patch'):
+                kw['data'] = params
+            else:
+                kw['params'] = params
+        meth = getattr(requests, method.lower())
+        if '://' not in url:
+            url = self.base_url + url
+        r = meth(url, **kw)
+        return FunctionalResponseWrapper(r)
 
 
 @pytest.fixture
@@ -864,10 +924,10 @@ class SimPyPIRequestHandler(httpserver.BaseHTTPRequestHandler):
                 start_response(200, headers)
                 self.wfile.write(b'\n'.join(releases))
                 return
-        elif p == ['', 'simple', '']:
+        elif p == ['', 'simple', ''] or p == ['', 'simple']:
             # root listing
             projects = [
-                '<a href="%s">%s</a>' % (k, v['title'])
+                '<a href="/simple/%s/">%s</a>' % (k, v['title'])
                 for k, v in simpypi.projects.items()]
             simpypi.add_log("do_GET", self.path, "found", list(simpypi.projects))
             start_response(200, headers)
@@ -941,48 +1001,77 @@ def call_devpi_in_dir():
     def devpi(server_dir, args):
         from devpi_server.main import main
         from _pytest.monkeypatch import MonkeyPatch
+        from _pytest.pytester import RunResult
         m = MonkeyPatch()
         m.setenv("DEVPI_SERVERDIR", server_dir)
         m.setattr("sys.argv", [devpiserver])
+        cap = py.io.StdCaptureFD()
+        cap.startall()
+        now = py.std.time.time()
         try:
             main(args)
         finally:
             m.undo()
+            out, err = cap.reset()
+            del cap
+        return RunResult(
+            0, out.split("\n"), err.split("\n"), py.std.time.time() - now)
 
     return devpi
 
 
-@pytest.yield_fixture(scope="module")
-def master_host_port(request, call_devpi_in_dir, server_directory):
+@pytest.fixture(scope="class")
+def master_serverdir(server_directory):
+    return server_directory.join("master")
+
+
+@pytest.yield_fixture(scope="class")
+def master_host_port(request, call_devpi_in_dir, master_serverdir):
     host = 'localhost'
     port = get_open_port(host)
-    master_dir = server_directory.join("master").strpath
+    args = [
+        "devpi-server",
+        "--serverdir", master_serverdir.strpath,
+        "--role", "master",
+        "--host", host,
+        "--port", str(port),
+        "--requests-only"]
+    if not master_serverdir.join('.nodeinfo').exists():
+        subprocess.check_call(
+            args + ["--init"])
+    p = subprocess.Popen(args)
+    try:
+        wait_for_port(host, port)
+        yield (host, port)
+    finally:
+        p.terminate()
+        p.wait()
+
+
+@pytest.fixture(scope="class")
+def replica_serverdir(server_directory):
+    return server_directory.join("replica")
+
+
+@pytest.yield_fixture(scope="class")
+def replica_host_port(request, call_devpi_in_dir, master_host_port, replica_serverdir):
+    host = 'localhost'
+    port = get_open_port(host)
+    args = [
+        "devpi-server", "--start",
+        "--host", host, "--port", str(port),
+        "--master-url", "http://%s:%s" % master_host_port]
+    if not replica_serverdir.join('.nodeinfo').exists():
+        args.append("--init")
     call_devpi_in_dir(
-        master_dir,
-        ["devpi-server", "--start", "--host", host, "--port", str(port)])
+        replica_serverdir.strpath,
+        args)
     try:
         wait_for_port(host, port)
         yield (host, port)
     finally:
         call_devpi_in_dir(
-            master_dir,
-            ["devpi-server", "--stop"])
-
-
-@pytest.yield_fixture(scope="module")
-def replica_host_port(request, call_devpi_in_dir, server_directory):
-    host = 'localhost'
-    port = get_open_port(host)
-    replica_dir = server_directory.join("replica").strpath
-    call_devpi_in_dir(
-        replica_dir,
-        ["devpi-server", "--start", "--host", host, "--port", str(port)])
-    try:
-        wait_for_port(host, port)
-        yield (host, port)
-    finally:
-        call_devpi_in_dir(
-            replica_dir,
+            replica_serverdir.strpath,
             ["devpi-server", "--stop"])
 
 
@@ -1015,9 +1104,12 @@ def _nginx_host_port(host, port, call_devpi_in_dir, server_directory):
 
     orig_dir = server_directory.chdir()
     try:
+        args = ["devpi-server", "--gen-config", "--host", host, "--port", str(port)]
+        if not server_directory.join('.nodeinfo').exists():
+            args.append("--init")
         call_devpi_in_dir(
             server_directory.strpath,
-            ["devpi-server", "--gen-config", "--host", host, "--port", str(port)])
+            args)
     finally:
         orig_dir.chdir()
     nginx_directory = server_directory.join("gen-config")
@@ -1044,12 +1136,12 @@ def _nginx_host_port(host, port, call_devpi_in_dir, server_directory):
     return (p, nginx_port)
 
 
-@pytest.yield_fixture(scope="module")
+@pytest.yield_fixture(scope="class")
 def nginx_host_port(request, call_devpi_in_dir, server_directory):
     if sys.platform.startswith("win"):
         pytest.skip("no nginx on windows")
     # we need the skip above before master_host_port is called
-    (host, port) = request.getfuncargvalue("master_host_port")
+    (host, port) = request.getfixturevalue("master_host_port")
     (p, nginx_port) = _nginx_host_port(
         host, port, call_devpi_in_dir, server_directory)
     try:
@@ -1059,7 +1151,7 @@ def nginx_host_port(request, call_devpi_in_dir, server_directory):
         p.wait()
 
 
-@pytest.yield_fixture(scope="module")
+@pytest.yield_fixture(scope="class")
 def nginx_replica_host_port(replica_host_port, call_devpi_in_dir, server_directory):
     (host, port) = replica_host_port
     (p, nginx_port) = _nginx_host_port(
@@ -1136,6 +1228,9 @@ class SimPyPI:
             info['length'] = length
         self.files[relpath] = info
 
+    def remove_file(self, relpath):
+        del self.files[relpath]
+
 
 @pytest.fixture
 def simpypi(simpypiserver):
